// IoT Brick Library (ESP32)
// 1.00 - 2017-06-10
// 1.03 - 2017-11-10
// 1.04 - 2017-12-14
// (c) Copyright 2017
//
// Zur Verwendung mit dem Allnet IoT32-Brick


#include <WiFi.h>
#include <TimeLib.h>
#include <NTPClient.h>
#include <WiFiUdp.h>
#include <DNSServer.h>
#include <WebServer.h>
#include <HTTPClient.h>
#include <WiFiManager.h> 
#include <EEPROM.h>
#include "SSD1306.h"                                         // Alias für #include "SSD1306Wire.h"
#include <driver/adc.h>
#include <driver/dac.h>

#include "all_Type.h"
#include "all_String.h"
#include "all_Terminal.h"
#include "all_HTTP.h"


// *** Globale Konstanten ***

#define    BRKclock_PIN 18                           // Für die BRK-Clock: ESP8266 = 14, ESP32 = 18
const bool t_ShowMessages               = true;      // Status-Mitteilungen anzeigen
const bool t_HideMessages               = false;     // Status-Mitteilungen nicht anzeigen
const int  IoT_Analog0                  = 36;        // ADC0 im ESP32 hat den Index 36
const int  IoT_Analog1                  = 37;        // ADC1 im ESP32 hat den Index 37
const int  IoT_Analog2                  = 38;        // ADC2 im ESP32 hat den Index 38
const int  IoT_Analog3                  = 39;        // ADC3 im ESP32 hat den Index 39


// *** Globale Variablen ***

TwoWire          IoT_Wire               = TwoWire(1);// Externer I2C-Bus (SCL 22, SDA 21)
SSD1306          IoT_Display(0x3c, 4, 15);           // Variable für das OLED Display (128 x 64 Pixel)
WiFiUDP          IoT_NTPupdate;                      // NTP Update
NTPClient        IoT_NTP(IoT_NTPupdate, "ptbtime1.ptb.de", 3600); // NTP Client, 1 Stunde Zeitoffset in Deutschland
byte             IoT_NTPcurrentEvent = 0;            // Wird beim ESP32 nicht benötigt

long             IoT_WLANstartSeconds   = -1;        // Zeitpunkt, an dem das WLAN erfolgreich verbunden wurde in ms.
String           IoT_WLANupDateTime     = "";        // Datum und Uhrzeit, an dem der NTP-Server erfolgreich verbunden wurde
bool             IoT_WLANinited         = false;     // Variable für den erfolgreichen (oder nicht) WLAN Connect
bool             IoT_OLEDinited         = false;     // Variable für die Initialisierung der OLED-Display Benutzung
bool             IoT_NTPinited          = false;     // Variable für die Initialisierung der NTP Benutzung
bool             IoT_NTPeventAvail      = false;     // Variable für das Auftreten eines NTP Events
long             IoT_EEPROMsize         = 0;         // Größe des verbauten EPROMs
byte             IoT_StaticIP           = 0;         // 0 = Dynamische IP-Adresse, 1 = Statische IP-Adresse
String           IoT_DeviceID           = "IoT-" + left(newuniqueid(), 8);  // Eindeutigen Namen erzeugen


// *** Variablen für verwendete DHCP-IP-Adressen ***

byte             IoT_dynamicIP_address0 = 0;         // 4 Bytes IP-Adresse
byte             IoT_dynamicIP_address1 = 0;
byte             IoT_dynamicIP_address2 = 0;
byte             IoT_dynamicIP_address3 = 0;
byte             IoT_dynamicIP_gateway0 = 0;         // 4 Bytes Gateway
byte             IoT_dynamicIP_gateway1 = 0;
byte             IoT_dynamicIP_gateway2 = 0;
byte             IoT_dynamicIP_gateway3 = 0;
byte             IoT_dynamicIP_subnet0  = 0;         // 4 Bytes Subnetzmaske
byte             IoT_dynamicIP_subnet1  = 0;
byte             IoT_dynamicIP_subnet2  = 0;
byte             IoT_dynamicIP_subnet3  = 0;


// *** Forward-Deklarationen ***

void   IoT_Idle ();
void   IoT_DisplayClear (int fontSize);
void   IoT_DisplayFontSize (int fontSize);
void   IoT_DisplayUpdate ();
void   IoT_DisplayDrawText (int x, int y, String text);
void   IoT_DisplayAlignText (OLEDDISPLAY_TEXT_ALIGNMENT alignment);
String IoT_NTPdatetime ();


// *** Initialisierung des IoT-Bricks ***

void IoT_Init (bool IoTbrick)                       // Standard-Initialisierung, muss als Erstes aufgerufen werden
{
  pinMode(0, INPUT);                                // Eingebauten Taster benutzbar machen
  switch (t_CPUtype())
  {
    case t_ESP32:
      Serial.begin(921600);
      IoT_EEPROMsize = 4096;
      EEPROM.begin(IoT_EEPROMsize);                 // Daten aus dem EEPROM auslesen
      break;
    case t_ArduinoMega2560:
      Serial.begin(921600);
      break;
    case t_Teensy31:
      Serial.begin(921600); // Beherrscht alle Baudraten, macht bis zu 14 MBit.
      break;
    default: // t_ArduinoNano, t_ArduinoDue u.a.
      Serial.begin(115200);
      break;
  }
  String clr = "   " + chr(13) + chr(27) + "*";     // Teminal-Bildschirm löschen
  Serial.print(clr);

  if (IoTbrick)
  {
    pinMode(16,OUTPUT);
    digitalWrite(16, LOW);                          // Set GPIO16 LOW, um das OLED zu resetten
    delay(50); 
    digitalWrite(16, HIGH);                         // Wenn das OLED läuft, muss GPIO16 auf HIGH gesetzt werden
    IoT_Display.init();                             // Initialisiert die OLED graphische Anzeige
    IoT_OLEDinited = true;                          // Variable für OLED-Initialisierung setzen
    IoT_Display.flipScreenVertically();             // Anzeige spiegeln (USB-Stecker ist auf der linken Seite)
    IoT_DisplayClear(10);                           // Eingebautes OLED-Display löschen

    pinMode(2, OUTPUT);                             // GPIO 02 auf Ausgabe schalten (CS)
    pinMode(5, OUTPUT);                             // GPIO 05 auf Ausgabe schalten (CSO)
    pinMode(17, OUTPUT);                            // GPIO 17 auf Ausgabe schalten (U2TXD)
    pinMode(18, OUTPUT);                            // GPIO 18 auf Ausgabe schalten (SCK)
    pinMode(19, OUTPUT);                            // GPIO 19 auf Ausgabe schalten (MISO)
    pinMode(23, OUTPUT);                            // GPIO 23 auf Ausgabe schalten (MOSI)

    digitalWrite(2, LOW);                           // GPIO 02 auf 0V schalten (CS)
    digitalWrite(5, LOW);                           // GPIO 05 auf 0V schalten (CSO)
    digitalWrite(17, LOW);                          // GPIO 17 auf 0V schalten (U2TXD)
    digitalWrite(18, LOW);                          // GPIO 18 auf 0V schalten (SCK)
    digitalWrite(19, LOW);                          // GPIO 19 auf 0V schalten (MISO)
    digitalWrite(23, LOW);                          // GPIO 23 auf 0V schalten (MOSI)

    pinMode(13, INPUT);                             // GPIO 13 auf Eingabe schalten (Taster 1)
    pinMode(34, INPUT);                             // GPIO 34 auf Eingabe schalten (Taster 3)
    pinMode(35, INPUT);                             // GPIO 35 auf Eingabe schalten (Taster 2)

    pinMode(IoT_Analog0, INPUT);                    // GPIO 36 auf Eingabe schalten (ADC 1.0)
    pinMode(IoT_Analog1, INPUT);                    // GPIO 37 auf Eingabe schalten (ADC 1.1)
    pinMode(IoT_Analog2, INPUT);                    // GPIO 38 auf Eingabe schalten (ADC 1.2)
    pinMode(IoT_Analog3, INPUT);                    // GPIO 39 auf Eingabe schalten (ADC 1.3)

    adc1_config_width(ADC_WIDTH_12Bit);             // 12 Bit Auflösung für ADC 1.0 bis 1.3
    adc1_config_channel_atten(ADC1_CHANNEL_0, ADC_ATTEN_11db);  // Vorverstärkung ADC 1.0 auf 11 dB
    adc1_config_channel_atten(ADC1_CHANNEL_1, ADC_ATTEN_11db);  // Vorverstärkung ADC 1.1 auf 11 dB
    adc1_config_channel_atten(ADC1_CHANNEL_2, ADC_ATTEN_11db);  // Vorverstärkung ADC 1.2 auf 11 dB
    adc1_config_channel_atten(ADC1_CHANNEL_3, ADC_ATTEN_11db);  // Vorverstärkung ADC 1.3 auf 11 dB
  }
}


// *** WLAN des IoT-Bricks ***

String IoT_WLANstatus ()
{
  switch (WiFi.status())
  {
    case 0:
      return ("Idle");
      break;
    case 1:
      return ("No SSID available");
      break;
    case 2:
      return ("Scan completed");
      break;
    case 3: // WL_CONNECTED
      return ("Connected");
      break;
    case 4:
      return ("Connect failed");
      break;
    case 5:
      return ("Connection lost");
      break;
    case 6:
      return ("Disconnected");
      break;
    default:
      return ("N/A");
      break;
  }
  IoT_Idle();
}

bool IoT_WLANconnected ()                                       // Ist eine WLAN-Verbindung erfolgreich hergestellt?
{
  return (WiFi.isConnected());
}

bool IoT_WLANconnect (String WLANssid, String WLANpassword)     // Funktion um die WLAN Verbindung aufzubauen
{                                                               // incl. Fortschrittsbalken
  IoT_WLANinited = false;
  int progress = 0;

  Serial.print(chr(12));

  // Da manche Access Points langsamer antworten als andere, muss man ein paar Sekunden auf die Antwort warten.
  // Solange WLAN nicht verbunden ist, Fortschrittsbalken im Display und Punkte in der Console anzeigen.
  while (not(IoT_WLANinited) && (progress < 100))
  {  
    if ((progress % 25) == 0)
    {
      if (progress == 50)                                       // Wenn nach 50 versuchen noch keine Wifi-Verbindung
      {                                                         // zu Stande kommt, dann alle Konfigurationsdaten
        Serial.println("");                                     // aus dem Flash-Speicher löschen, beim ESP32 nicht möglich
        //ESP.eraseConfig(); 
        Serial.print("Resetting Configuration. ");
      }
      if ((progress == 0) || (progress == 50))
      {
        Serial.println("WLAN access point " + WLANssid);
        Serial.print("Connecting ");
      }
      IoT_Idle();
      WiFi.persistent(false);                                   // Räumt mit alten RF-Daten auf und sorgt dafür, dass
      WiFi.mode(WIFI_OFF);                                      // zuverlässig eine WiFi-Verbindung aufgebaut wird,
      WiFi.mode(WIFI_STA);                                      // auch dann, wenn WiFi erst später startet. WiFi erst
      WiFi.begin(WLANssid.c_str(), WLANpassword.c_str());       // ausschalten und dann wieder einschalten.
      delay(500);                                               // 500 ms. warten
    }
    IoT_WLANinited = (IoT_WLANstatus() == "Connected");
    //Serial.println(String(progress) + " Status = " + IoT_WLANstatus() + " = " + boolstr(IoT_WLANinited, "True", "False"));
    Serial.print(".");
    
    if (IoT_OLEDinited)
    {
      IoT_Display.clear();
      IoT_Display.drawProgressBar(0, 32, 120, 10, progress);
      IoT_Display.setTextAlignment(TEXT_ALIGN_CENTER);
      IoT_Display.drawString(64, 15, String(progress) + "%");
      IoT_Display.display();
    }
    IoT_Idle();

    if (not(IoT_WLANinited))                // Verbindung noch nicht hergestellt?
    {
      delay(500);                           // 500 ms. warten
      progress++;                           // Durchgang um 1 erhöhen
    }
  }
  Serial.println("");
  if (IoT_WLANinited) // Falls Verbindung erfolgreich, zeige WLAN Namen und die DHCP IP Adresse in der Console
  {
    Serial.print("Connected to ");
    Serial.print(WiFi.SSID());
    Serial.print(", IP: ");
    Serial.println(WiFi.localIP());
    IoT_dynamicIP_address0 = WiFi.localIP()[0];                 // 4 Bytes IP-Adresse
    IoT_dynamicIP_address1 = WiFi.localIP()[1];
    IoT_dynamicIP_address2 = WiFi.localIP()[2];
    IoT_dynamicIP_address3 = WiFi.localIP()[3];
    IoT_dynamicIP_gateway0 = WiFi.gatewayIP()[0];               // 4 Bytes Gateway
    IoT_dynamicIP_gateway1 = WiFi.gatewayIP()[1];
    IoT_dynamicIP_gateway2 = WiFi.gatewayIP()[2];
    IoT_dynamicIP_gateway3 = WiFi.gatewayIP()[3];
    IoT_dynamicIP_subnet0  = WiFi.subnetMask()[0];              // 4 Bytes Subnetzmaske
    IoT_dynamicIP_subnet1  = WiFi.subnetMask()[1];
    IoT_dynamicIP_subnet2  = WiFi.subnetMask()[2];
    IoT_dynamicIP_subnet3  = WiFi.subnetMask()[3];
    IoT_WLANstartSeconds   = long(millis() / 1000);             // Zeitpunkt in Sekunden, an dem das WLAN gestartet wurde (Überlauf nach 50 Tagen)
  }
  else // gebe Fehlermeldung in der Console aus, falls Verbindung fehlgeschlagen
  {
    Serial.println("Connection failed");
  }
  
  return (IoT_WLANinited);
}

String IoT_WLANaddress (bool leadingZero) // Lokale IP-Adresse des WLAN-Netzwerkes (wahlweise mit führenden Nullen)
{
  return (strIP(WiFi.localIP(), leadingZero));
}

String IoT_WLANgateway (bool leadingZero) // Gateway (Router) IP-Adresse des WLAN-Netzwerkes (wahlweise mit führenden Nullen)
{
  return (strIP(WiFi.gatewayIP(), leadingZero));
}

String IoT_WLANsubnet (bool leadingZero) // Subnetzmaske des WLAN-Netzwerkes (wahlweise mit führenden Nullen)
{
  return (strIP(WiFi.subnetMask(), leadingZero));
}

String IoT_WLANssid () // Name des WLAN-Netzwerkes
{
  return (WiFi.SSID());
}

int IoT_WLANrssi () // Signalstärke in dBm (immer ein negativer Wert, z.B. -70 dBm)
{
  return (WiFi.RSSI());
}

String IoT_WLANuptime (bool germanLanguage)                         // Dauer der aktuellen WLAN-Verbindung
{
  long t = long(millis() / 1000) - IoT_WLANstartSeconds;            // Dauer in Sekunden
  int s = t % 60;
  t /= 60;
  int m = t % 60;
  t /= 60;
  int h = t % 24;
  t /= 24;
  String result = str(t);
  if (germanLanguage)
  {
    result += " Tage ";                                             // Zeitdauer im Format "T Tage hh:mm:ss"
  }
  else
  {
    result += " days ";                                             // Zeitdauer im Format "T days hh:mm:ss"
  }
  result += strform(h, 48, 2, false);
  result += ":";
  result += strform(m, 48, 2, false);
  result += ":";
  result += strform(s, 48, 2, false);
  return (result);                                                  // Zeitdauer im Format "T days hh:mm:ss"
}

String IoT_WLANstartDatetime ()                                     // Startzeit der aktuellen WLAN-Verbindung
{
  return (IoT_WLANupDateTime);                                      // Datum & Uhrzeit im Format tt.mm.jjjj hh:mm:ss
}

void IoT_WLANautoConnect (String devName, bool useSavedSettings)    // Verbindung mit WiFi-Manager herstellen
{
  WiFiManager wifiManager;
  IoT_DeviceID = devName;
  if (IoT_DeviceID == "")
  {
    IoT_DeviceID = "IoT-" + left(newuniqueid(), 8);
  }
  if (not(useSavedSettings))
  {
    Serial.println("Erasing old WiFi-Configuration...");
    wifiManager.resetSettings();                                    // Funktioniert auf dem ESP32 nicht
    wifiManager.autoConnect(IoT_DeviceID.c_str());                  // Verbindung aufbauen, muss vor dem WiFi.disconnect(true) stehen
    WiFi.disconnect(true);                                          // Sorgt auf dem ESP32 dafür, dass die alten Einstellungen gelöscht werden
    delay(1000);                                                    // Zwingend notwendig, damit ein erneuter Verbindungsaufbau klappt
  }
  if (IoT_OLEDinited)
  {
    IoT_DisplayClear(16);                                           // OLED Display löschen (16 Punkt Schrift)
    IoT_DisplayAlignText(TEXT_ALIGN_CENTER);
    IoT_DisplayDrawText(63, 0, "WLAN wählen:");
    IoT_DisplayFontSize(16);
    IoT_DisplayDrawText(63, 25, IoT_DeviceID);
    IoT_DisplayFontSize(10);
    IoT_DisplayDrawText(63, 50, "Danach HotSpot wählen");
    IoT_DisplayUpdate();                                            // Display aktualisieren
    IoT_DisplayClear(16);                                           // OLED Display löschen (16 Punkt Schrift)
  }
  wifiManager.autoConnect(IoT_DeviceID.c_str());                    // Verbindung ggf. mit gespeicherten Einstellungen 
  if (IoT_OLEDinited)
  {
    IoT_DisplayUpdate();                                            // Display aktualisieren
  }
  Serial.print("Connected to ");
  Serial.print(WiFi.SSID());
  Serial.print(", IP: ");
  Serial.println(WiFi.localIP());
  IoT_WLANinited = true;
  IoT_dynamicIP_address0 = WiFi.localIP()[0];                       // 4 Bytes IP-Adresse
  IoT_dynamicIP_address1 = WiFi.localIP()[1];
  IoT_dynamicIP_address2 = WiFi.localIP()[2];
  IoT_dynamicIP_address3 = WiFi.localIP()[3];
  IoT_dynamicIP_gateway0 = WiFi.gatewayIP()[0];                     // 4 Bytes Gateway
  IoT_dynamicIP_gateway1 = WiFi.gatewayIP()[1];
  IoT_dynamicIP_gateway2 = WiFi.gatewayIP()[2];
  IoT_dynamicIP_gateway3 = WiFi.gatewayIP()[3];
  IoT_dynamicIP_subnet0  = WiFi.subnetMask()[0];                    // 4 Bytes Subnetzmaske
  IoT_dynamicIP_subnet1  = WiFi.subnetMask()[1];
  IoT_dynamicIP_subnet2  = WiFi.subnetMask()[2];
  IoT_dynamicIP_subnet3  = WiFi.subnetMask()[3];
  IoT_WLANstartSeconds   = long(millis() / 1000);                   // Zeitpunkt in Sekunden, an dem das WLAN gestartet wurde (Überlauf nach 50 Tagen)
}

void IoT_WLANautoConnect (bool useSavedSettings)                    // Verbindung mit WiFi-Manager herstellen
{
  IoT_WLANautoConnect("", useSavedSettings);
}


// *** OLED-Display Funktionen ***

void IoT_DisplayFontSize (int fontSize)           // Schriftgröße setzen
{
  if (IoT_OLEDinited)
  {
    switch (fontSize)
    {
      case 10:
        IoT_Display.setFont(ArialMT_Plain_10);      // Schriftart setzen. Hiermit sind 6 Zeilen Text möglich
        break;
      case 16:
        IoT_Display.setFont(ArialMT_Plain_16);      // Schriftart setzen. Hiermit sind 4 Zeilen Text möglich
        break;
      case 24:
        IoT_Display.setFont(ArialMT_Plain_24);      // Schriftart setzen. Hiermit sind 2 Zeilen Text möglich
        break;
      default:
        IoT_Display.setFont(ArialMT_Plain_10);      // Schriftart setzen. Hiermit sind 6 Zeilen Text möglich
        break;
    }
  }
  IoT_Idle();
}

void IoT_DisplayClear (int fontSize)                // Bildschirm löschen und Schriftgröße setzen
{
  if (IoT_OLEDinited)
  {
    IoT_Display.clear();
    IoT_Display.setTextAlignment(TEXT_ALIGN_LEFT);  // Linksbündige Darstellung des Textes
    IoT_DisplayFontSize(fontSize);
  }
}

void IoT_DisplayUpdate ()                           // Bildschirm aktualisieren
{
  if (IoT_OLEDinited)
  {
    IoT_Display.display();
  }
  IoT_Idle();
}

void IoT_DisplayDrawText (int x, int y, String text) // Text an der angegebenen Koordinate schreiben
{
  if (IoT_OLEDinited)
  {
    IoT_Display.drawString(x, y, text);
  }
  IoT_Idle();
}

void IoT_DisplayAlignText (OLEDDISPLAY_TEXT_ALIGNMENT alignment)
{
  if (IoT_OLEDinited)
  {
    IoT_Display.setTextAlignment(alignment);
  }
}

void IoT_DisplayWLANstatus ()
{
  if (IoT_OLEDinited)
  {
    IoT_DisplayClear(10);                                           // OLED Display löschen (10 Punkt Schrift)
    String state = IoT_WLANstatus();                                // Verbindungsstatus
    if (IoT_WLANinited)                                             // Falls WLAN Verbindung erfolgreich, Status & IP im OLED Display anzeigen
    {
      IoT_DisplayDrawText(5, 0, "WLAN: " + IoT_WLANssid());         // WLAN Netzwerk Name
      if (state == "Connected")                                     // Wenn erfolgreich verbunden, dann Signalstärke hinzufügen
      {
        state = state + " (" + str(IoT_WLANrssi()) + "dBm)";        // Signalstärke
      }
      IoT_DisplayDrawText(5, 10, state);                            // Verbindungsstatus & Signalstärke
      IoT_DisplayDrawText(5, 20, "IP: " + IoT_WLANaddress(true));   // IP Adresse
      IoT_DisplayDrawText(5, 30, "GW: " + IoT_WLANgateway(true));   // IP Gateway
      IoT_DisplayDrawText(5, 40, "NT: " + IoT_WLANsubnet(true));    // IP Subnetzmaske
    }
    else                                                            // Falls WLAN Verbindung gescheitert ist, Status im OLED Display anzeigen
    {
      IoT_DisplayDrawText(5, 10, state); 
    }
  }
}


// *** NTP Funktionen ***

bool IoT_NTPinit ()                               // Funktion um die NTP Verbindung aufzubauen
{
  if (IoT_WLANinited)                             // Nur wenn WLAN verbunden ist
  {
    IoT_NTP.begin();                              // Verbindung zu NTP Zeit-Server herstellen
    if (IoT_NTP.forceUpdate());
    {
      setTime(IoT_NTP.getEpochTime());
      if (IoT_WLANupDateTime == "")
      {
        IoT_WLANupDateTime = "(invalid)";
        IoT_WLANupDateTime = IoT_NTPdatetime();
      }
    }
    IoT_NTPeventAvail = true;
    IoT_NTPinited = true;
  }
  else
  {
    IoT_NTPinited = false;                        // Ohne WLAN ist kein NTP möglich
  }
  IoT_Idle();
  return (IoT_NTPinited);
}

String IoT_NTPtime ()
{
  if (IoT_NTP.update());
  {
    setTime(IoT_NTP.getEpochTime());
    if ((IoT_WLANupDateTime == "") || (left(IoT_WLANupDateTime, 10) == "01.01.1970"))
    {
      IoT_WLANupDateTime = "(invalid)";
      IoT_WLANupDateTime = IoT_NTPdatetime();
      IoT_NTPeventAvail = true;
    }
  }
  String s = strform(hour(), 48, 2, false);
  s += ":";
  s += strform(minute(), 48, 2, false);
  s += ":";
  s += strform(second(), 48, 2, false);
  return (s);                                     // Uhrzeit im Format hh:mm:ss
}

String IoT_NTPdate ()
{
  if (IoT_NTP.update());
  {
    setTime(IoT_NTP.getEpochTime());
    if ((IoT_WLANupDateTime == "") || (left(IoT_WLANupDateTime, 10) == "01.01.1970"))
    {
      IoT_WLANupDateTime = "(invalid)";
      IoT_WLANupDateTime = IoT_NTPdatetime();
      IoT_NTPeventAvail = true;
    }
  }
  String s = strform(day(), 48, 2, false);
  s += ".";
  s += strform(month(), 48, 2, false);
  s += ".";
  s += strform(year(), 48, 4, false);
  return (s);                                     // Datum im Format tt.mm.jjjj           
}

String IoT_NTPdatetime ()
{
  String s = IoT_NTPdate();
  s += " ";
  s += IoT_NTPtime();
  return (s);                                     // Datum & Uhrzeit im Format tt.mm.jjjj hh:mm:ss
}

bool IoT_NTPvalid ()
{
  bool valid = IoT_NTP.update();
  return (IoT_NTP.getEpochTime() > 1000000000);   // Zeit darf nicht weit in der Vergangenheit liegen
}

void IoT_NTPprintEvent (byte ignoredParameter)
{
  if (IoT_NTPvalid())                                             // Fehlermeldung überprüfen
  {
    Serial.println("Got NTP time: " + IoT_NTPdatetime());
  }
  else                                                            // Zeit wurde erfolgreich empfangen
  {
    Serial.println("Time Sync error: NTP server not reachable");
  }
}

String IoT_NTPdaylightSavingTime (bool germanLanguage)
{
  bool summertime = false;
  if (month() == 3)
  {
    if (day() >= 26)
    {
      summertime = true;
    }
  }
  else if (month() == 10)
  {
    if (day() <= 28)
    {
      summertime = true;
    }
  }
  else if ((month() > 3) && (month() < 10))
  {
    summertime = true;
  }
  if (germanLanguage)
  {
    return (summertime ? "Sommerzeit" : "Winterzeit");
  }
  else
  {
    return (summertime ? "Summer Time" : "Winter Time");
  }
}


// *** EEPROM-Befehle für den IoT-Brick ***

void IoT_EEPROMclear ()                          // Die Konfiguration des WiFi-Managers wird hiervon nicht berührt
{
  bool success = true;
  const byte zero = 0;
  for (int i = 0; i <= IoT_EEPROMsize; i++)
  {
    EEPROM.put(i, zero);
  }
}

void IoT_EEPROMupdate ()
{
  EEPROM.commit();                               // Daten in das EEPROM zurückschreiben
}

String IoT_EEPROMgetString (int addr, bool messages)
{
  if ((addr >= 0) && (addr <= 4095))                         // EEPROM 0...4095 = $0000...0FFF
  {
    byte L = 0;
    EEPROM.get(addr, L);                                     // Länge des Strings aus dem EEPROM laden
    if (messages)
    {
      Serial.print("Load " + str(L) + " chars from EEPROM $" + hexword(addr) + ": ");
    }
    if (L == 0)
    {
      return ("");
    }
    else
    {
      String s = spc(L);
      for (int i = 0; i < L; i++)
      {
        byte zchn = 0;
        EEPROM.get(addr + 1 + i, zchn);                      // Buchstabe für Buchstabe laden
        s[i] = zchn;
        if (messages)
        {
          Serial.print(chr(zchn));
        }
      }
      if (messages)
      {
        Serial.println("");
      }
      return (s);
    }
  }
  else
  {
    return ("");
  }
}

String IoT_EEPROMgetString (int addr)
{
  return (IoT_EEPROMgetString(addr, false));
}

bool IoT_EEPROMputString (int addr, String s, bool messages)
{
  if ((addr >= 0) && (addr <= 4095))                         // EEPROM 0...4095 = $0000...0FFF
  {
    if (s.length() > 255)
    {
      return (false);
    }
    else
    {
      byte L = len(s);
      if (messages)
      {
        Serial.print("Save " + str(L) + " chars to EEPROM $" + hexword(addr) + ": ");
      }
      EEPROM.put(addr, L);                                   // Länge des Strings im EEPROM sichern
      if (L > 0)
      {
        for (int i = 0; i < L; i++)
        {
          byte zchn = s[i];
          EEPROM.put(addr + 1 + i, zchn);                    // Buchstabe für Buchstabe sichern
          if (messages)
          {
            Serial.print(chr(zchn));
          }
        }
      }
      if (messages)
      {
        Serial.println("");
      }
      return (true);
    }
  }
  else
  {
    return (false);
  }
}

bool IoT_EEPROMputString (int addr, String s)
{
  return (IoT_EEPROMputString(addr, s, false));
}


// *** Utilites für den IoT-Brick ***

void IoT_Idle ()                                  // Verhindert das Festfrieren und Neustarten bei komplexen Aufgaben
{
//ESP.wdtFeed();                                  // Watchdog Timer zurücksetzen
//delay(1);                                       // Alternativ kann man auch 1 ms. warten, kostet aber CPU-Zeit
}

void IoT_WatchDog (bool active)                   // Schaltet den IoT-Brick-Watchdog ein (true) oder aus (false)
{
}

bool IoT_Keypress (byte GPIO)                     // Eingebauten Taster abfragen
{
  return ((digitalRead(GPIO) == LOW));
}

bool IoT_Keypress ()                              // Eingebauten Taster abfragen
{
  return (IoT_Keypress(0));
}

void IoT_WaitMessage (byte GPIO)
{
    IoT_DisplayClear(16);                         // OLED Display löschen (16 Punkt Schrift)
    IoT_DisplayAlignText(TEXT_ALIGN_CENTER);      // Zentrierte Darstellung des Textes
    if (GPIO == 0)
    {
      IoT_DisplayDrawText(64,  5, "Den Taster am");
    }
    else
    {
      IoT_DisplayDrawText(64,  5, "Den Taster " + String(GPIO) + " am");
    }
    IoT_DisplayDrawText(64, 25, "IoT-Brick drücken");
    IoT_DisplayDrawText(64, 45, "um fortzufahren.");
    IoT_DisplayUpdate();
}

void IoT_WaitMessage ()
{
    IoT_WaitMessage(0);                           // Eingebauter Taster
}

void IoT_WaitKeypress (byte GPIO, uint64 timeout, bool message)
{
  long startTime = millis();
  if (timeout == 0)                               // Timeout 0 bedeutet: Kein Timeout
  {
    timeout = 20000000000000;                     // 20 Milliarden Sekunden = mehr als 630 Jahre
  }
  if (message)
  {
    IoT_WaitMessage (GPIO);
  }
  while (not(IoT_Keypress(GPIO)) && (abs(millis() - startTime) < timeout))
  {                                               // Wenn Taster nicht gedrückt wird und noch kein Timeout
    delay(1);                                     // 1 ms. warten
    IoT_WebServer.handleClient();                 // Bediene die http Anfragen
  }
  if (message)
  {
    IoT_DisplayClear(16);                         // OLED Display löschen (16 Punkt Schrift)
    IoT_DisplayUpdate();
  }
}

void IoT_WaitKeypress (uint64 timeout, bool message)
{
  IoT_WaitKeypress(0, timeout, message);          // Eingebauter Taster
}

void IoT_WaitNoKeypress (byte GPIO)
{
  while (IoT_Keypress(GPIO))                      // Warten, bis der Taster losgelassen wurde
  {
    delay(1);
    IoT_Idle();
    IoT_WebServer.handleClient();                 // Bediene die http Anfragen
  }
}

void IoT_WaitNoKeypress ()
{
  IoT_WaitNoKeypress(0);
}

void IoT_TerminalWaitInit (bool showMessage)
{
  if (showMessage == t_ShowMessages)
  {
    if (IoT_OLEDinited)
    {
      IoT_DisplayClear(16);
      IoT_DisplayAlignText(TEXT_ALIGN_CENTER);
      IoT_DisplayDrawText(63, 2, "Bitte mit");
      IoT_DisplayDrawText(63, 22, "Terminal");
      IoT_DisplayDrawText(63, 42, "verbinden");
      IoT_DisplayUpdate();
    }
  }

  t_TerminalRespond = "";
  while (t_TerminalWidth == 0)
  {
    t_TerminalInit();
    if (t_TerminalWidth == 0)
    {
      delay(500); // 500 ms. warten
    }
  }
  t_TerminalRespond = "";

  if (showMessage == t_ShowMessages)
  {
    if (IoT_OLEDinited)
    {
      IoT_DisplayClear(10);
      IoT_DisplayAlignText(TEXT_ALIGN_CENTER);
      IoT_DisplayDrawText(63, 0, "Verbindung");
      IoT_DisplayDrawText(63, 12, "ist hergestellt");
      IoT_DisplayDrawText(63, 24, "Protokoll:");
      IoT_DisplayDrawText(63, 36, t_TerminalType);
      IoT_DisplayDrawText(63, 48, String(t_TerminalWidth) + " x " + String(t_TerminalHeight) + " Zeichen");
      IoT_DisplayUpdate();
    }
  }
}

void IoT_ShutDown ()
{
  if (IoT_OLEDinited)
  {
    IoT_Display.clear();                         // Display löschen
    IoT_Display.display();                       // Display aktualisieren
  }
  if (IoT_EEPROMsize > 0)
  {
    EEPROM.end();                                // Daten in das EEPROM zurückschreiben und Speicherplatz freigeben
  }
  while (true)
  {
    IoT_Idle();                                  // Watchdog Timer zurücksetzen
    t_TerminalGetKey();
    delay(100);                                  // 100 ms. warten
  }
}

int IoT_AnalogRead (int channel)
{
  switch (channel)
  {
    case IoT_Analog0:
      return (adc1_get_voltage(ADC1_CHANNEL_0));
      break;
    case IoT_Analog1:
      return (adc1_get_voltage(ADC1_CHANNEL_1));
      break;
    case IoT_Analog2:
      return (adc1_get_voltage(ADC1_CHANNEL_2));
      break;
    case IoT_Analog3:
      return (adc1_get_voltage(ADC1_CHANNEL_3));
      break;
    default:
      return (0);
      break;
  }
}

void IoT_AnalogWrite (int channel, byte value)
{
  switch (channel)
  {
    case IoT_Analog1:
      dac_out_voltage(DAC_CHANNEL_1, value);
      break;
    case IoT_Analog2:
      dac_out_voltage(DAC_CHANNEL_2, value);
      break;
    case 25:
      dac_out_voltage(DAC_CHANNEL_1, value);
      break;
    case 26:
      dac_out_voltage(DAC_CHANNEL_2, value);
      break;
    default:
      break;
  }
}
